<?xml version="1.0" encoding="UTF-8"?>
<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
  xmlns:xs="http://www.w3.org/2001/XMLSchema"
  xmlns:m="http://www.w3.org/1998/Math/MathML"
  xmlns:mine="mine:local"
  exclude-result-prefixes="xs m mine"
  version="3.0">

  <!-- parameter -->
  <xsl:param name="element" as="xs:string"/>
  <xsl:param name="path" as="xs:string?"/>
  <xsl:param name="content" as="xs:string"/>
  
  <!-- include some utility functions -->
  <xsl:include href="util-indent.xsl"/>
  <xsl:include href="util-insert.xsl"/>


  <!-- TOP-LEVEL PASS - baseline identity transform -->
  <xsl:mode on-no-match="shallow-copy"/>

  <!-- TOP-LEVEL PASS - apply multiple processing passes to input document -->
  <xsl:template match="/*" priority="100">
    <xsl:variable name="result" as="element()">
      <xsl:next-match/>  <!-- apply #default mode templates first -->
    </xsl:variable>
    <xsl:apply-templates select="$result" mode="indent-stuff"/>  <!-- process/remove @indent directives -->
  </xsl:template>


  <!-- insert an <indexterm> into /topic/prolog/metadata/keywords -->
  <xsl:template match="*[local-name(.) eq $element]">
    <!-- apply other templates first so we chain with other templates -->
    <xsl:variable name="result" as="element()">
      <xsl:next-match/>  <!-- apply other templates first -->
    </xsl:variable>

    <!-- insert the content -->
    <xsl:apply-templates select="$result" mode="insert-stuff">
      <xsl:with-param name="path" as="xs:string*" select="$path"/>  <!-- element path the content should be inserted into (empty string for current element) -->
      <xsl:with-param name="content" as="element()*" select="parse-xml($content)/node()" tunnel="yes"/>  <!-- content to insert -->
    </xsl:apply-templates>
  </xsl:template>

</xsl:stylesheet>
